import os
import json
import yaml
import logging
import signal
import psutil
from .. import setlib
from .. import dbhandler

log = logging.getLogger(__name__)


def submit(args):
    dirs = setlib.get_dirs(args.test)
    settings = setlib.get_settings(dirs.user_config_dir, dirs.user_data_dir)
    queue = settings["queue"]

    assert os.path.exists(args.payload)
    assert os.path.isfile(args.payload)

    log.debug(f"attempting to open payload at '{args.payload}'")
    with open(args.payload, "r") as infile:
        if args.payload.endswith("json"):
            payload = json.load(infile)
        elif args.payload.endswith("yml") or args.payload.endswith("yaml"):
            payload = yaml.full_load(infile)
    pstr = json.dumps(payload)
    log.info("queueing 'payload' into 'queue'")
    dbhandler.queue_payload(queue["path"], pstr, type=queue["type"])


def status(args):
    dirs = setlib.get_dirs(args.test)
    settings = setlib.get_settings(dirs.user_config_dir, dirs.user_data_dir)
    state = settings["state"]
    queue = settings["queue"]

    if args.jobid == "state":
        pips = dbhandler.pipeline_get_all(state["path"], type=state["type"])
        print(
            f"{'pipeline':20s} {'ready':5s} {'jobid':5s} {'(PID)':9s} {'sampleid':20s} "
        )
        print("=" * 65)
        for pip in pips:
            sampleid, ready, jobid, pid = dbhandler.pipeline_get_info(
                state["path"], pip, state["type"]
            )
            rstr = "yes" if ready else "no"
            job = f"{str(jobid):5s} ({pid})" if jobid is not None else str(jobid)
            print(f"{pip:20s} {rstr:5s} {job:15s} {str(sampleid):20s}")
    elif args.jobid == "queue":
        jobs = dbhandler.job_get_all(queue["path"], type=queue["type"])
        running = dbhandler.pipeline_get_running(state["path"], type=state["type"])
        print(f"{'jobid':6s} {'status':6s} {'pid':7s} {'pipeline':20s}")
        print("=" * 42)
        for jobid, payload, status in jobs:
            if status.startswith("q"):
                print(f"{str(jobid):6s} {status}")
            elif status.startswith("r"):
                for pip, pjobid, pid in running:
                    if pjobid == jobid:
                        print(f"{str(jobid):6s} {status:6s} {str(pid):7s} {pip:20s}")
    else:
        jobid = int(args.jobid)
        ji = dbhandler.job_get_info(queue["path"], jobid, type=queue["type"])
        payload, status, sub, exe, com = ji
        print(f"job '{jobid}':")
        print(f"  - status: '{status}'")
        print(f"  - submitted at: '{sub}'")
        if status.startswith("r"):
            print(f"  - running since:'{exe}'")
            running = dbhandler.pipeline_get_running(state["path"], type=state["type"])
            for pip, pjobid, pid in running:
                if pjobid == jobid:
                    print(f"  - on pipeline: '{pip}'")
                    print(f"  - with PID: '{pid}'")
                    break
        if status.startswith("c"):
            print(f"  - executed at:  '{exe}'")
            print(f"  - completed at: '{com}'")


def cancel(args):
    dirs = setlib.get_dirs(args.test)
    settings = setlib.get_settings(dirs.user_config_dir, dirs.user_data_dir)
    state = settings["state"]
    queue = settings["queue"]
    jobid = int(args.jobid)
    jobinfo = dbhandler.job_get_info(queue["path"], jobid, type=queue["type"])
    status = jobinfo[1]
    log.debug(f"found job {jobid} with status '{status}'")
    if status in {"q", "qw"}:
        log.info(f"setting job {jobid} to status 'cd'")
        dbhandler.job_set_status(queue["path"], "cd", jobid, type=queue["type"])
    elif status == "r":
        running = dbhandler.pipeline_get_running(state["path"], type=state["type"])
        for pip, pjobid, pid in running:
            if pjobid == jobid:
                log.warning(f"cancelling a running job {jobid} with pid {pid}")
                proc = psutil.Process(pid=pid)
                cproc = proc.children()
                for p in cproc:
                    p.send_signal(signal.SIGTERM)
                    log.debug(f"SIGTERM sent to pid {p.pid}")


def load(args):
    dirs = setlib.get_dirs(args.test)
    settings = setlib.get_settings(dirs.user_config_dir, dirs.user_data_dir)
    state = settings["state"]

    log.debug(f"checking whether pipeline '{args.pipeline}' exists.")
    pips = dbhandler.pipeline_get_all(state["path"], type=state["type"])
    assert args.pipeline in pips, f"pipeline '{args.pipeline}' not found."

    log.info(f"loading sample '{args.sample}' into pipeline '{args.pipeline}'")
    dbhandler.pipeline_load_sample(
        state["path"], args.pipeline, args.sample, type=state["type"]
    )


def eject(args):
    dirs = setlib.get_dirs(args.test)
    settings = setlib.get_settings(dirs.user_config_dir, dirs.user_data_dir)
    state = settings["state"]

    log.debug(f"checking whether pipeline '{args.pipeline}' exists.")
    pips = dbhandler.pipeline_get_all(state["path"], type=state["type"])
    assert args.pipeline in pips, f"pipeline '{args.pipeline}' not found."

    log.debug(f"checking if pipeline '{args.pipeline}' is running.")
    sampleid, ready, jobid, pid = dbhandler.pipeline_get_info(
        state["path"], args.pipeline, state["type"]
    )

    if sampleid is not None:
        assert jobid is None and pid is None, (
            "cannot remove a sample" " from a running pipeline"
        )
        log.info(f"ejecting sample '{sampleid}' from pipeline '{args.pipeline}'")
        dbhandler.pipeline_eject_sample(
            state["path"], args.pipeline, type=state["type"]
        )
    else:
        log.info(f"pipeline '{args.pipeline}' is already empty")


def ready(args):
    dirs = setlib.get_dirs(args.test)
    settings = setlib.get_settings(dirs.user_config_dir, dirs.user_data_dir)
    state = settings["state"]

    log.debug(f"checking whether pipeline '{args.pipeline}' exists.")
    pips = dbhandler.pipeline_get_all(state["path"], type=state["type"])
    assert args.pipeline in pips, f"pipeline '{args.pipeline}' not found."

    log.debug(f"checking if pipeline '{args.pipeline}' is running.")
    sampleid, ready, jobid, pid = dbhandler.pipeline_get_info(
        state["path"], args.pipeline, state["type"]
    )

    if jobid is None and pid is None:
        log.info(f"marking pipeline '{args.pipeline}' as ready.")
        dbhandler.pipeline_reset_job(state["path"], args.pipeline, True, state["type"])
    else:
        log.warning(f"cannot mark pipeline as ready: job '{jobid}' is running.")



""" Bio-Logic OEM package data types.

This module provides a transcription of the EcLib DLL data types and constants.

As the names and values can be found in the Development Package documentation,
one should refer to the PDF for further explanations.

The main datatypes the module relies on are ctypes Structure, Enum and dataclass.

Note that the DLL has alignement requirements that can be achieved with
the _pack_ attribute in the Structures.

"""

from math import nan
from enum import Enum, EnumMeta
from dataclasses import dataclass

from .c_utils import *

# ==============================================================================#

# Max number of slots
MAX_SLOT_NB = 16

# Array of channels
ChannelsArray = c_bool * MAX_SLOT_NB

# Array of results
ResultsArray = c_int32 * MAX_SLOT_NB

DataBuffer = c_uint32 * 1000

# ==============================================================================#


@dataclass
class USB_device:

    index: int
    instrument: str
    serial: str

    def __str__(self):
        en_clair = f"{self.address} : {self.instrument} s/n {self.serial}"
        return en_clair

    @property
    def address(self):
        address = f"USB{self.index}"
        return address


# ==============================================================================#


@dataclass
class Ethernet_device:

    config: tuple
    instrument: str
    serial: str
    identifier: str
    name: str

    def __str__(self):
        en_clair = f"Ethernet{self.config} : {self.instrument}, s/n '{self.serial}'"
        if self.identifier:
            en_clair += f", id='{self.identifier}'"
        if self.name:
            en_clair += f", name='{self.name}'"
        return en_clair


# ==============================================================================#


class DEVICE(Enum):
    VMP = 0
    VMP2 = 1
    MPG = 2
    BISTAT = 3
    MCS_200 = 4
    VMP3 = 5
    VSP = 6
    HCP803 = 7
    EPP400 = 8
    EPP4000 = 9
    BISTAT2 = 10
    FCT150S = 11
    VMP300 = 12
    SP50 = 13
    SP150 = 14
    FCT50S = 15
    SP300 = 16
    CLB500 = 17
    HCP1005 = 18
    CLB2000 = 19
    VSP300 = 20
    SP200 = 21
    MPG2 = 22
    SP100 = 23
    MOSLED = 24
    KINEXXX = 25
    BCS815 = 26
    SP240 = 27
    MPG205 = 28
    MPG210 = 29
    MPG220 = 30
    MPG240 = 31
    BP300 = 32
    VMP3E = 33
    VSP3E = 34
    SP50E = 35
    SP150E = 36
    UNKNOWN = 255


# ------------------------------------------------------------------------------#

VMP3_FAMILY = (
    "VMP2",
    "VMP3",
    "BISTAT",
    "BISTAT2",
    "MCS_200",
    "VSP",
    "SP50",
    "SP150",
    "FCT50S",
    "FCT150S",
    "CLB500",
    "CLB2000",
    "HCP803",
    "HCP1005",
    "MPG2",
    "MPG205",
    "MPG210",
    "MPG220",
    "MPG240",
    "VMP3E",
    "VSP3E",
    "SP50E",
    "SP150E",
)

# ------------------------------------------------------------------------------#

VMP300_FAMILY = (
    "SP100",
    "SP200",
    "SP300",
    "VSP300",
    "VMP300",
    "SP240",
    "BP300",
)

# ------------------------------------------------------------------------------#


class DeviceInfo(POD):
    _fields_ = [
        ("DeviceCode", c_int32),
        ("RAMSize", c_int32),
        ("CPU", c_int32),
        ("NumberOfChannels", c_int32),
        ("NumberOfSlots", c_int32),
        ("FirmwareVersion", c_int32),
        ("FirmwareDate_yyyy", c_int32),
        ("FirmwareDate_mm", c_int32),
        ("FirmwareDate_dd", c_int32),
        ("HTdisplayOn", c_int32),
        ("NbOfConnectedPC", c_int32),
    ]


DEVICE_INFO = POINTER(DeviceInfo)

# ------------------------------------------------------------------------------#


class CHANNEL_BOARD(Enum):
    C340_IF0 = 0
    C340_IF2_Z = 1
    C340_OTHERS = 2
    C340_IF2_NONZ = 3
    C340_IF3_Z = 4
    C340_IF3_NONZ = 5
    C340_IF3_ZZ = 6
    C340_IF3_NZZ = 7
    C340_SP50 = 8
    C340_SP150Z = 9
    C340_SP150NZ = 10
    C437_Z = 11
    C437_NZ = 12
    C437_SP150Z = 13
    C437_SP150NZ = 14
    C437_MPG2 = 15
    C437_MPG2Z = 16
    C437_MPGX = 17
    C437_MPGXZ = 18
    C437_VMP3EZ = 19
    C437_VMP3ENZ = 20


class ChannelInfo(POD):
    _fields_ = [
        ("Channel", c_int32),  # Channel (0..15)
        ("BoardVersion", c_int32),  # Board version
        ("BoardSerialNumber", c_int32),  # Board serial number
        ("FirmwareCode", c_int32),  # Identifier of the firmware loaded on the channel
        ("FirmwareVersion", c_int32),  # Firmware version
        ("XilinxVersion", c_int32),  # Xilinx version
        ("AmpCode", c_int32),  # Amplifier code
        ("NbAmps", c_int32),  # Number of amplifiers (0..16)
        ("Lcboard", c_int32),  # Low current presence
        ("Zboard", c_int32),  # Impedance capabilities
        ("MUXboard", c_int32),  # MEA mux
        ("GPRAboard", c_int32),  # Analog ramp
        ("MemSize", c_int32),  # Memory size (in bytes)
        ("MemFilled", c_int32),  # Memory filled (in bytes)
        ("State", c_int32),  # Channel state : run/stop/pause
        ("MaxIRange", c_int32),  # Maximum I range allowed
        ("MinIRange", c_int32),  # Minimum I range allowed
        ("MaxBandwidth", c_int32),  # Maximum bandwidth allowed
        ("NbOfTechniques", c_int32),  # Number of techniques loaded
    ]


CH_INFO = POINTER(ChannelInfo)

# ------------------------------------------------------------------------------#


class HardwareConf(POD):
    _fields_ = [
        ("Connection", c_int32),  # Electrode connection
        ("Mode", c_int32),  # Channel mode
    ]


HW_CONF = POINTER(HardwareConf)


class HW_CNX(Enum):
    STANDARD = 0  # Standard connection
    CE_TO_GND = 1  # CE to ground connection
    WE_TO_GND = 2  # WE to ground connection
    HIGH_VOLTAGE = 3  # 48V connection


class HW_MODE(Enum):
    GROUNDED = 0  # Grounded mode
    FLOATING = 1  # Floating mode


# ------------------------------------------------------------------------------#


class OPTION_ERROR(Enum):
    NO_ERROR = 0  # No error found
    OPT_CHANGE = 1  # Number of options changed
    OPEN_IN = 2  # Open-in signal was asserted
    IRCMP_OVR = 3  # R Compensation overflow
    OPT_4A = 100  # 4A amplifier unknown error
    OPT_4A_OVRTEMP = 101  # 4A amplifier temperature overflow
    OPT_4A_BADPOW = 102  # 4A amplifier bad power
    OPT_4A_POWFAIL = 103  # 4A amplifier power fail
    OPT_48V = 200  # 48V amplifier unknown error
    OPT_48V_OVRTEMP = 201  # 48V amplifier temperature overflow
    OPT_48V_BADPOW = 202  # 48V amplifier bad power
    OPT_48V_POWFAIL = 203  # 48V amplifier power fail
    OPT_10A5V_ERR = 300  # 10A 5V amplifier error
    OPT_10A5V_OVRTEMP = 301  # 10A 5V amplifier overheat
    OPT_10A5V_BADPOW = 302  # 10A 5V amplifier bad power
    OPT_10A5V_POWFAIL = 303  # 10A 5V amplifier power fail


# -----------------------------------------------------------------------------#


class FIRMWARE(Enum):
    NONE = 0
    INTERPR = 1
    UNKNOWN = 4
    KERNEL = 5
    INVALID = 8
    ECAL = 10
    ECAL4 = 11


# ------------------------------------------------------------------------------#


class CurrentValues(POD):
    _fields_ = [
        ("State", c_int32),
        ("MemFilled", c_int32),
        ("TimeBase", c_float),
        ("Ewe", c_float),
        ("EweRangeMin", c_float),
        ("EweRangeMax", c_float),
        ("Ece", c_float),
        ("EceRangeMin", c_float),
        ("EceRangeMax", c_float),
        ("Eoverflow", c_int32),
        ("I", c_float),
        ("IRange", c_int32),
        ("Ioverflow", c_int32),
        ("ElapsedTime", c_float),
        ("Freq", c_float),
        ("Rcomp", c_float),
        ("Saturation", c_int32),
        ("OptErr", c_int32),
        ("OptPos", c_int32),
    ]


CURRENT_VALUES = POINTER(CurrentValues)

# ------------------------------------------------------------------------------#


class DataInfo(POD):
    _pack_ = 4
    _fields_ = [
        ("IRQskipped", c_int32),
        ("NbRows", c_int32),
        ("NbCols", c_int32),
        ("TechniqueIndex", c_int32),
        ("TechniqueID", c_int32),
        ("ProcessIndex", c_int32),
        ("loop", c_int32),
        ("StartTime", c_double),
        ("MuxPad", c_int32),
    ]


DATA_INFO = POINTER(DataInfo)

# ------------------------------------------------------------------------------#


class PARAM_TYPE(Enum):
    PARAM_INT = 0
    PARAM_BOOLEAN = 1
    PARAM_SINGLE = 2


# ------------------------------------------------------------------------------#


class EccParam(POD):
    _fields_ = [
        ("ParamStr", 64 * c_byte),
        ("ParamType", c_int32),
        ("ParamVal", c_uint32),
        ("ParamIndex", c_int32),
    ]


ECC_PARM = POINTER(EccParam)

# ------------------------------------------------------------------------------#


class EccParams(POD):
    _pack_ = 4
    _fields_ = [
        ("len", c_int32),
        ("pParams", ECC_PARM),
    ]


ECC_PARMS = POINTER(EccParams)


def ECC_PARM_ARRAY(nb):
    array_type = nb * EccParam
    return array_type()


# ------------------------------------------------------------------------------#


class TechniqueInfos(POD):
    _fields_ = [
        ("Id", c_int32),
        ("indx", c_int32),
        ("nbParams", c_int32),
        ("nbSettings", c_int32),
        ("Params", ECC_PARM),
        ("HardSettings", ECC_PARM),
    ]


TECHNIQUE_INFOS = POINTER(TechniqueInfos)

# ------------------------------------------------------------------------------#


class PROG_STATE(Enum):
    STOP = 0  # Channel is stopped
    RUN = 1  # Channel is running
    PAUSE = 2  # Channel is paused
    SYNC = 3  # grouped channels synchronization (stack)


# ------------------------------------------------------------------------------#


class I_RANGE(Enum):
    I_RANGE_KEEP = -1  # Keep previous
    I_RANGE_100pA = 0  # 100 pA
    I_RANGE_1nA = 1  # 1 nA VMP3
    I_RANGE_10nA = 2  # 10 nA VMP3
    I_RANGE_100nA = 3  # 100 nA VMP3
    I_RANGE_1uA = 4  # 1 μA VMP3
    I_RANGE_10uA = 5  # 10 μA VMP3
    I_RANGE_100uA = 6  # 100 μA VMP3
    I_RANGE_1mA = 7  # 1 mA VMP3
    I_RANGE_10mA = 8  # 10 mA VMP3
    I_RANGE_100mA = 9  # 100 mA VMP3
    I_RANGE_1A = 10  # 1 A VMP3
    I_RANGE_BOOSTER = 11  # Booster VMP3
    I_RANGE_AUTO = 12  # Auto range VMP3


# ------------------------------------------------------------------------------#


class E_RANGE(Enum):
    E_RANGE_2_5V = 0  # ±2.5V
    E_RANGE_5V = 1  # ±5V
    E_RANGE_10V = 2  # ±10V
    E_RANGE_AUTO = 3  # auto


# ------------------------------------------------------------------------------#


class BANDWIDTH(Enum):
    BW_1 = 1
    BW_2 = 2
    BW_3 = 3
    BW_4 = 4
    BW_5 = 5
    BW_6 = 6
    BW_7 = 7
    BW_8 = 8
    BW_9 = 9


# ------------------------------------------------------------------------------#


class GAIN(Enum):
    GAIN_1 = 0
    GAIN_10 = 1
    GAIN_100 = 2
    GAIN_1000 = 3


# ------------------------------------------------------------------------------#


class FILTER(Enum):
    FILTER_NONE = 0
    FILTER_50KHZ = 1
    FILTER_1KHZ = 2
    FILTER_5HZ = 3


# ------------------------------------------------------------------------------#


class AMPLIFIER(Enum):
    AMPL_NONE = 0  # No Amplifier
    AMPL_2A = 1  # Amplifier 2 A
    AMPL_1A = 2  # Amplifier 1 A
    AMPL_5A = 3  # Amplifier 5 A
    AMPL_10A = 4  # Amplifier 10 A
    AMPL_20A = 5  # Amplifier 20 A
    AMPL_HEUS = 6  # reserved
    AMPL_LC = 7  # Low current amplifier
    AMPL_80A = 8  # Amplifier 80 A
    AMPL_4AI = 9  # Amplifier 4 A
    AMPL_PAC = 10  # Fuel Cell Tester
    AMPL_4AI_VSP = 11  # Amplifier 4 A (VSP instrument)
    AMPL_LC_VSP = 12  # Low current amplifier (VSPinstrument)
    AMPL_UNDEF = 13  # Undefined amplifier
    AMPL_MUIC = 14  # reserved
    AMPL_ERROR = 15  # No amplifier (error case)
    AMPL_8AI = 16  # Amplifier 8 A
    AMPL_LB500 = 17  # Amplifier LB500
    AMPL_100A5V = 18  # Amplifier 100 A
    AMPL_LB2000 = 19  # Amplifier LB2000
    AMPL_1A48V = 20  # Amplifier 1A 48V
    AMPL_4A14V = 21  # Amplifier 4A 14V
    AMPL_5A_MPG2B = 22  # Amplifier 5A
    AMPL_10A_MPG2B = 23  # Amplifier 10A
    AMPL_20A_MPG2B = 24  # Amplifier 20A
    AMPL_40A_MPG2B = 25  # Amplifier 40A
    AMPL_COIN_CELL_HOLDER = 26  # Coin cell holder amplifier
    AMPL4_10A5V = 27  # Amplifier 10A 5V
    AMPL4_2A30V = 28  # Amplifier 2A 30V
    AMPL4_30A50V = 77  # 30A/50V amplifier
    AMPL3_50A60V = 93  # 50A/60V amplifier (FlexP 0160)
    AMPL3_200A12V = 97  # 200A/12V amplifier (FlexP 0012)
    AMPL3_50A60VII = 101  # 50A/60V amplifier (FlexP 0060)
    AMPL4_1A48VPII = 105  # 1A/48VP v2
    AMPL4_1A48VPIII = 129  # 1A/48VP v3


# ------------------------------------------------------------------------------#


class ERROR(Enum):
    NOERROR = 0  # "No error"
    GEN_NOTCONNECTED = -1  # "No instrument connected"
    GEN_CONNECTIONINPROGRESS = -2  # "Connection in progress"
    GEN_CHANNELNOTPLUGGED = -3  # "Selected channel(s) unplugged"
    GEN_INVALIDPARAMETERS = -4  # "Invalid function parameters"
    GEN_FILENOTEXISTS = -5  # "Selected file does not exist"
    GEN_FUNCTIONFAILED = -6  # "Function failed"
    GEN_NOCHANNELSELECTED = -7  # "No channel selected"
    GEN_INVALIDCONF = -8  # "Invalid instrument configuration"
    GEN_ECLAB_LOADED = -9  # "EC-Lab firmware loaded on the instrument"
    GEN_LIBNOTCORRECTLYLOADED = -10  # "Library not correctly loaded in memory"
    GEN_USBLIBRARYERROR = -11  # "USB library not correctly loaded in memory"
    GEN_FUNCTIONINPROGRESS = -12  # "Function already in progress"
    GEN_CHANNEL_RUNNING = -13  # "Selected channel(s) already used"
    GEN_DEVICE_NOTALLOWED = -14  # "Device not allowed"
    GEN_UPDATEPARAMETERS = -15  # "Invalid update function parameters"
    INSTR_VMEERROR = -101  # "Internal instrument communication failed"
    INSTR_TOOMANYDATA = (
        -102
    )  # "Too many data to transfer from the instrument (device error)"
    INSTR_RESPNOTPOSSIBLE = -103  # "Selected channel(s) unplugged (device error)"
    INSTR_RESPERROR = -104  # "Instrument response error"
    INSTR_MSGSIZEERROR = -105  # "Invalid message size"
    COMM_COMMFAILED = -200  # "Communication failed with the instrument"
    COMM_CONNECTIONFAILED = -201  # "Cannot establish connection with the instrument"
    COMM_WAITINGACK = -202  # "Waiting for the instrument response"
    COMM_INVALIDIPADDRESS = -203  # "Invalid IP address"
    COMM_ALLOCMEMFAILED = -204  # "Cannot allocate memory in the instrument"
    COMM_LOADFIRMWAREFAILED = -205  # "Cannot load firmware into selected channel(s)"
    COMM_INCOMPATIBLESERVER = -206  # "Communication firmware not compatible"
    COMM_MAXCONNREACHED = -207  # "Maximum number of allowed connections reached"
    FIRM_FIRMFILENOTEXISTS = -300  # "Cannot find kernel.bin file"
    FIRM_FIRMFILEACCESSFAILED = -301  # "Cannot read kernel.bin file"
    FIRM_FIRMINVALIDFILE = -302  # "Invalid kernel.bin file"
    FIRM_FIRMLOADINGFAILED = -303  # "Cannot load kernel.bin on the selected channel(s)"
    FIRM_XILFILENOTEXISTS = -304  # "Cannot find FPGA file"
    FIRM_XILFILEACCESSFAILED = -305  # "Cannot read FPGA file"
    FIRM_XILINVALIDFILE = -306  # "Invalid FPGA file"
    FIRM_XILLOADINGFAILED = -307  # "Cannot load FPGA file on the selected channel(s)"
    FIRM_FIRMWARENOTLOADED = -308  # "No firmware loaded on the selected channel(s)"
    FIRM_FIRMWAREINCOMPATIBLE = (
        -309
    )  # "Loaded firmware not compatible with the library"
    TECH_ECCFILENOTEXISTS = -400  # "Cannot find the selected ECC file"
    TECH_INCOMPATIBLEECC = -401  # "ECC file not compatible with the channel firmware"
    TECH_ECCFILECORRUPTED = -402  # "ECC file corrupted"
    TECH_LOADTECHNIQUEFAILED = -403  # "Cannot load the ECC file"
    TECH_DATACORRUPTED = -404  # "Data returned by the instrument are corrupted"
    TECH_MEMFULL = -405  # "Cannot load techniques: full memory"
    OPT_CHANGE = 1  # "Number of options changed"
    OPT_OPEN_IN = 2  # "Open-in signal was asserted"
    OPT_4A_ERROR = 100  # 4A10V amplifier error
    OPT_4A_OVERTEMP = 101  # 4A10V amplifier overload temperature
    OPT_4A_BADPOWER = 102  # 4A10V amplifier invalid power
    OPT_4A_POWERFAIL = 103  # 4A10V amplifier power fail
    OPT_48V_ERROR = 200  # 1A48V amplifier error
    OPT_48V_OVERTEMP = 201  # 1A48V amplifier overload temperature
    OPT_48V_BADPOWER = 202  # 1A48V amplifier invalid power
    OPT_48V_POWERFAIL = 203  # 1A48V amplifier power fail
    OPT_10A5V_ERROR = 300  # 10A5V amplifier error
    OPT_10A5V_OVERTEMP = 301  # 10A5V amplifier overload temperature
    OPT_10A5V_BADPOWER = 302  # 10A5V amplifier invalid power
    OPT_10A5V_POWERFAIL = 303  # 10A5V amplifier power fail
    OPT_1A48VP_ERROR = 600  # 1A48VP amplifier error
    OPT_1A48VP_OVERTEMP = 601  # 1A48VP amplifier overheat
    OPT_1A48VP_BADPOWER = 602  # 1A48VP amplifier bad power
    OPT_1A48VP_POWERFAIL = 603  # 1A48VP amplifier power fail


# ==============================================================================#


class FIND_ERROR(Enum):
    NO_ERROR = 0  # "No error"
    UNKNOWN_ERROR = -1  # unknown error
    INVALID_PARAMETER = -2  # invalid function parameters
    ACK_TIMEOUT = -10  # instrument response timeout
    EXP_RUNNING = -11  # experiment is running on instrument
    CMD_FAILED = -12  # instrument do not execute command
    FIND_FAILED = -20  # find failed
    SOCKET_WRITE = (
        -21
    )  # cannot write the request of the descriptions of Ethernet instruments
    SOCKET_READ = -22  # cannot read descriptions of Ethernet instrument
    CFG_MODIFY_FAILED = -30  # set TCP/IP parameters failed
    READ_PARAM_FAILED = -31  # deserialization of TCP/IP parameters failed
    EMPTY_PARAM = -32  # not any TCP/IP parameters in serialization
    IP_FORMAT = -33  # invalid format of IP address
    NM_FORMAT = -34  # invalid format of netmask address
    GW_FORMAT = -35  # invalid format of gateway address
    IP_NOT_FOUND = -38  # instrument to modify not found
    IP_ALREADYEXIST = -39  # new IP address in TCP/IP parameters


# ==============================================================================#



from typing import Any, Callable
import importlib
import time
import multiprocessing
import os
import json
from datetime import datetime, timezone
import logging

from .logger_funcs import log_listener_config, log_listener, log_worker_config


def driver_api(
    driver: str,
    command: str,
    jobqueue: multiprocessing.Queue,
    logger: logging.Logger,
    address: str,
    channel: int,
    **kwargs: dict,
) -> Any:
    m = importlib.import_module(f"tomato.drivers.{driver}")
    func = getattr(m, command)
    return func(address, channel, jobqueue, logger, **kwargs)


def data_poller(
    driver: str,
    jq: multiprocessing.Queue,
    lq: multiprocessing.Queue,
    address: str,
    channel: int,
    device: str,
    root: str,
    kwargs: dict,
) -> None:
    log_worker_config(lq)
    log = logging.getLogger()
    pollrate = kwargs.pop("pollrate", 10)
    verbose = bool(kwargs.pop("verbose", 0))
    log.debug(f"in 'data_poller', {pollrate=}")
    cont = True
    while cont:
        ts, done, metadata = driver_api(
            driver, "get_status", jq, log, address, channel, **kwargs
        )
        if verbose:
            isots = datetime.fromtimestamp(ts, tz=timezone.utc).isoformat()
            isots = isots.replace(":", "")
            fn = os.path.join(root, f"{device}_{isots}_status.json")
            log.debug(f"'writing status info into '{fn}'")
            with open(fn, "w") as of:
                json.dump(metadata, of)
        ts, nrows, data = driver_api(
            driver, "get_data", jq, log, address, channel, **kwargs
        )
        while nrows > 0:
            isots = datetime.fromtimestamp(ts, tz=timezone.utc).isoformat()
            isots = isots.replace(":", "")
            fn = os.path.join(root, f"{device}_{isots}_data.json")
            log.debug(f"found {nrows} data rows, writing into '{fn}'")
            with open(fn, "w") as of:
                json.dump(data, of)
            ts, nrows, data = driver_api(
                driver, "get_data", jq, log, address, channel, **kwargs
            )
        if done:
            cont = False
        else:
            time.sleep(pollrate)
    log.info(f"rejoining main thread")
    return


def driver_worker(
    settings: dict, pipeline: dict, payload: dict, jobid: int, logfile: str
) -> None:

    jq = multiprocessing.Queue(maxsize=0)

    log = logging.getLogger(__name__)
    log.debug("starting 'log_listener'")
    lq = multiprocessing.Queue(maxsize=0)
    listener = multiprocessing.Process(
        target=log_listener,
        name="log_listener",
        args=(lq, log_listener_config, logfile),
    )
    listener.start()
    log.debug(f"started 'log_listener' on pid {listener.pid}")

    root = os.path.join(settings["queue"]["storage"], str(jobid))
    jobs = []
    for vi, v in enumerate(pipeline["devices"]):
        log.info(f"device id: {vi+1} out of {len(pipeline['devices'])}")
        log.info(f"{vi+1}: processing device '{v['tag']}' of type '{v['driver']}'")
        drv, addr, ch, tag = v["driver"], v["address"], v["channel"], v["tag"]
        dpar = settings["drivers"].get(drv, {})
        pl = payload["method"][tag]
        smpl = payload["sample"]

        log.debug(f"{vi+1}: getting status")
        ts, ready, metadata = driver_api(drv, "get_status", jq, log, addr, ch, **dpar)
        assert ready, f"Failed: device '{tag}' is not ready."

        log.debug(f"{vi+1}: starting payload")
        start_ts = driver_api(
            drv, "start_job", jq, log, addr, ch, **dpar, payload=pl, **smpl
        )
        metadata["uts"] = start_ts

        log.debug(f"{vi+1}: writing metadata")
        isots = datetime.fromtimestamp(ts, tz=timezone.utc).isoformat().replace(":", "")
        fn = os.path.join(root, f"{tag}_{isots}_status.json")
        with open(fn, "w") as of:
            json.dump(metadata, of)
        kwargs = dpar
        kwargs.update(
            {
                "pollrate": v.get("pollrate", 10),
                "verbose": v.get("verbose", 0),
            }
        )
        log.info(f"{vi+1}: starting 'data_poller': every {kwargs['pollrate']}s")
        p = multiprocessing.Process(
            name=f"data_poller_{jobid}_{tag}",
            target=data_poller,
            args=(drv, jq, lq, addr, ch, tag, root, kwargs),
        )
        jobs.append(p)
        p.start()
        log.info(f"{vi+1}: started 'data_poller' on pid {p.pid}")

    log.info("waiting for all 'data_poller' jobs to join")
    log.info("------------------------------------------")
    ret = None
    for p in jobs:
        p.join()
        if p.exitcode == 0:
            log.info(f"'data_poller' with pid {p.pid} closed successfully")
        else:
            log.critical(f"'data_poller' with pid {p.pid} was terminated")
            ret = 1

    log.info("-----------------------")
    log.info("quitting 'log_listener'")
    lq.put_nowait(None)
    listener.join()
    jq.close()
    return ret


def driver_reset(
    settings: dict,
    pipeline: dict,
) -> None:
    log = logging.getLogger(__name__)
    for vi, v in enumerate(pipeline["devices"]):
        log.info(f"device id: {vi+1} out of {len(pipeline['devices'])}")
        log.info(f"{vi+1}: processing device '{v['tag']}' of type '{v['driver']}'")
        drv, addr, ch, tag = v["driver"], v["address"], v["channel"], v["tag"]
        dpar = settings["drivers"].get(drv, {})

        log.debug(f"{vi+1}: resetting device")
        driver_api(drv, "stop_job", addr, ch, **dpar)

        log.debug(f"{vi+1}: getting status")
        ts, ready, metadata = driver_api(drv, "get_status", addr, ch, **dpar)
        assert ready, f"Failed: device '{tag}' is not ready."



""" Bio-Logic OEM package technique constants.

This module provides an enumeration of technique identifiers which DLL calls return.

"""

from enum import Enum

# =============================================================================#


class TECH_ID(Enum):
    NONE = 0  # None
    OCV = 100  # Open Circuit Voltage (Rest)
    CA = 101  # Chrono-amperometry
    CP = 102  # Chrono-potentiometry
    CV = 103  # Cyclic Voltammetry
    PEIS = 104  # Potentio Electrochemical Impedance Spectroscopy
    POTPULSE = 105  # (unused)
    GALPULSE = 106  # (unused)
    GEIS = 107  # Galvano Electrochemical Impedance Spectroscopy
    STACKPEIS_SLAVE = 108  # Potentio Electrochemical Impedance Spectroscopy on stack
    STACKPEIS = 109  # Potentio Electrochemical Impedance Spectroscopy on stack
    CPOWER = 110  # Constant Power
    CLOAD = 111  # Constant Load
    FCT = 112  # (unused)
    SPEIS = 113  # Staircase Potentio Electrochemical Impedance Spectroscopy
    SGEIS = 114  # Staircase Galvano Electrochemical Impedance Spectroscopy
    STACKPDYN = 115  # Potentio dynamic on stack
    STACKPDYN_SLAVE = 116  # Potentio dynamic on stack
    STACKGDYN = 117  # Galvano dynamic on stack
    STACKGEIS_SLAVE = 118  # Galvano Electrochemical Impedance Spectroscopy on stack
    STACKGEIS = 119  # Galvano Electrochemical Impedance Spectroscopy on stack
    STACKGDYN_SLAVE = 120  # Galvano dynamic on stack
    CPO = 121  # (unused)
    CGA = 122  # (unused)
    COKINE = 123  # (unused)
    PDYN = 124  # Potentio dynamic
    GDYN = 125  # Galvano dynamic
    CVA = 126  # Cyclic Voltammetry Advanced
    DPV = 127  # Differential Pulse Voltammetry
    SWV = 128  # Square Wave Voltammetry
    NPV = 129  # Normal Pulse Voltammetry
    RNPV = 130  # Reverse Normal Pulse Voltammetry
    DNPV = 131  # Differential Normal Pulse Voltammetry
    DPA = 132  # Differential Pulse Amperometry
    EVT = 133  # Ecorr vs. time
    LP = 134  # Linear Polarization
    GC = 135  # Generalized corrosion
    CPP = 136  # Cyclic Potentiodynamic Polarization
    PDP = 137  # Potentiodynamic Pitting
    PSP = 138  # Potentiostatic Pitting
    ZRA = 139  # Zero Resistance Ammeter
    MIR = 140  # Manual IR
    PZIR = 141  # IR Determination with Potentiostatic Impedance
    GZIR = 142  # IR Determination with Galvanostatic Impedance
    LOOP = 150  # Loop (used for linked techniques)
    TO = 151  # Trigger Out
    TI = 152  # Trigger In
    TOS = 153  # Trigger Set
    CPLIMIT = 155  # Chrono-potentiometry with limits
    GDYNLIMIT = 156  # Galvano dynamic with limits
    CALIMIT = 157  # Chrono-amperometry with limits
    PDYNLIMIT = 158  # Potentio dynamic with limits
    LASV = 159  # Large amplitude sinusoidal voltammetry
    MP = 167  # Modular Pulse
    CASG = 169  # Constant amplitude sinusoidal micro galvano polarization
    CASP = 170  # Constant amplitude sinusoidal micro potentio polarization


# =============================================================================#



import logging
import logging.handlers


def log_listener_config(path: str):
    root = logging.getLogger()
    h = logging.FileHandler(path, mode="a")
    f = logging.Formatter("%(asctime)s:%(levelname)-8s:%(processName)-10s:%(message)s")
    h.setFormatter(f)
    root.addHandler(h)


def log_listener(queue, configurer, path):
    configurer(path)
    while True:
        record = queue.get()
        if record is None:
            break
        logger = logging.getLogger(record.name)
        logger.handle(record)


def log_worker_config(queue):
    h = logging.handlers.QueueHandler(queue)
    root = logging.getLogger()
    root.addHandler(h)
    root.setLevel(logging.DEBUG)



import logging
import multiprocessing

from datetime import datetime, timezone

from .kbio_wrapper import (
    get_kbio_techpath,
    payload_to_ecc,
    parse_raw_data,
    get_kbio_api,
)


def get_status(
    address: str,
    channel: int,
    jobqueue: multiprocessing.Queue,
    logger: logging.Logger,
    dllpath: str = None,
    **kwargs: dict,
) -> tuple[float, dict]:
    """
    Get the current status of the device.

    Parameters
    ----------
    address
        IP address of the potentiostat.

    channel
        Numeric, 1-indexed ID of the channel.

    dllpath
        Path to the BioLogic DLL file.

    Returns
    -------
    timestamp, ready, metadata: tuple[float, bool, dict]
        Returns a tuple containing the timestamp, readiness status, and
        associated metadata.

    """
    api = get_kbio_api(dllpath)
    metadata = {}
    metadata["dll_version"] = api.GetLibVersion()
    logger.debug(f"connecting to '{address}:{channel}'")
    id_, device_info = api.Connect(address)
    metadata["device_model"] = device_info.model
    metadata["device_channels"] = device_info.NumberOfChannels
    channel_info = api.GetChannelInfo(id_, channel)
    dt = datetime.now(timezone.utc)
    metadata["channel_state"] = channel_info.state
    metadata["channel_board"] = channel_info.board
    metadata["channel_amp"] = channel_info.amplifier if channel_info.NbAmps else None
    metadata["channel_I_ranges"] = [channel_info.min_IRange, channel_info.max_IRange]
    logger.debug(f"disconnecting from '{address}:{channel}'")
    api.Disconnect(id_)
    if metadata["channel_state"] in ["STOP"]:
        ready = True
    else:
        ready = False
    return dt.timestamp(), ready, metadata


def get_data(
    address: str,
    channel: int,
    jobqueue: multiprocessing.Queue,
    logger: logging.Logger,
    dllpath: str = None,
    **kwargs: dict,
) -> tuple[float, dict]:
    """
    Get cached data from the device.

    Parameters
    ----------
    address
        IP address of the potentiostat.

    channel
        Numeric, 1-indexed ID of the channel.

    dllpath
        Path to the BioLogic DLL file.

    Returns
    -------
    timestamp, nrows, data: tuple[float, int, dict]
        Returns a tuple containing the timestamp and associated metadata.

    """
    api = get_kbio_api(dllpath)
    logger.debug(f"connecting to '{address}:{channel}'")
    id_, device_info = api.Connect(address)
    logger.debug(f"getting data")
    data = api.GetData(id_, channel)
    dt = datetime.now(timezone.utc)
    logger.debug(f"disconnecting from '{address}:{channel}'")
    api.Disconnect(id_)
    data = parse_raw_data(api, data, device_info.model)
    return dt.timestamp(), data["technique"]["data_rows"], data


def start_job(
    address: str,
    channel: int,
    jobqueue: multiprocessing.Queue,
    logger: logging.Logger,
    payload: list[dict],
    dllpath: str = None,
    capacity: float = 0.0,
    **kwargs: dict,
) -> float:
    """
    Start a job on the device.

    The function first translates the ``payload`` into an instrument-specific
    language, using the ``capacity`` provided if necessary. The converted
    ``payload`` is then submitted to the device, overwriting any current job
    information.

    Parameters
    ----------
    address
        IP address of the potentiostat.

    channel
        Numeric, 1-indexed ID of the channel.

    dllpath
        Path to the BioLogic DLL file.

    payload
        A protocol describing the techniques to be executed and their order.

    capacity
        The capacity information for the studied battery cell. Only required for
        battery-testing applications or for payloads where currents are specified
        using C or D rates.

    Returns
    -------
    timestamp
        A timestamp corresponding to the start of the job execution.

    """
    api = get_kbio_api(dllpath)
    logger.debug("translating payload to ECC")
    eccpars = payload_to_ecc(api, payload, capacity)
    ntechs = len(eccpars)
    first = True
    last = False
    ti = 1
    logger.debug(f"connecting to '{address}:{channel}'")
    id_, device_info = api.Connect(address)
    for techname, pars in eccpars:
        if ti == ntechs:
            last = True
        techfile = get_kbio_techpath(dllpath, techname, device_info.model)
        logger.debug(f"loading technique {ti}: '{techname}'")
        api.LoadTechnique(
            id_, channel, techfile, pars, first=first, last=last, display=False
        )
        ti += 1
        first = False
    logger.debug(f"starting run on '{address}:{channel}'")
    api.StartChannel(id_, channel)
    dt = datetime.now(timezone.utc)
    logger.info(f"run started at '{dt}'")
    logger.debug(f"disconnecting from '{address}:{channel}'")
    api.Disconnect(id_)
    return dt.timestamp()


def stop_job(
    address: str,
    channel: int,
    jobqueue: multiprocessing.Queue,
    logger: multiprocessing.Queue,
    dllpath: str,
    **kwargs: dict,
) -> float:
    """
    Stop a job running on the device.

    This function stops any currently running technique on the specified channel
    of the device. No data is returned.

    Parameters
    ----------
    address
        IP address of the potentiostat.

    channel
        Numeric, 1-indexed ID of the channel.

    dllpath
        Path to the BioLogic DLL file.

    Returns
    -------
    timestamp
        A timestamp corresponding to the start of the job execution.

    """
    api = get_kbio_api(dllpath)
    logger.debug(f"connecting to '{address}:{channel}'")
    id_, device_info = api.Connect(address)
    logger.debug(f"stopping run on '{address}:{channel}'")
    api.StopChannel(id_, channel)
    dt = datetime.now(timezone.utc)
    logger.info(f"run stopped at '{dt}'")
    api.Disconnect(id_)
    jobqueue.close()
    return dt.timestamp()



import os
from typing import Union

import logging

log = logging.getLogger(__name__)

from .kbio.kbio_api import KBIO_api
from .kbio.kbio_types import PROG_STATE, VMP3_FAMILY, EccParams
from .kbio.tech_types import TECH_ID
from .kbio.kbio_tech import make_ecc_parm, make_ecc_parms, ECC_parm
from .kbio.c_utils import c_is_64b

from .tech_params import named_params, techfiles, datatypes, I_ranges, E_ranges


def get_test_magic(
    variable: str, sign: str, logic: str = "or", active: bool = True
) -> int:
    magic = int(active)

    assert logic.lower() in {"and", "or"}
    magic += 2 * int(logic.lower() in {"and"})

    assert sign.lower() in {"max", "min"}
    magic += 4 * int(sign.lower() in {"max"})

    assert variable.lower() in {"voltage", "E", "current", "I"}
    magic += 96 * int(variable.lower() in {"current", "I"})

    return magic


def get_num_steps(tech: dict) -> int:
    ns = 1
    for k in {
        "current",
        "voltage",
        "is_delta",
        "time",
        "scan_rate" "limit_voltage_min",
        "limit_voltage_max",
        "limit_current_min",
        "limit_current_max",
    }:
        if k in tech and isinstance(tech[k], list):
            ns = max(len(tech[k]), ns)
    return ns


def pad_steps(param: Union[list, int, float], ns: int) -> list:
    if isinstance(param, list):
        ret = [0] * ns
        for i in range(len(param)):
            ret[i] = param[i]
    else:
        ret = [param] * ns
    return ret


def _current(val: Union[list, str, float], capacity: float) -> float:
    if isinstance(val, float):
        return val
    elif "/" in val:
        pre, post = val.split("/")
        if pre == "C":
            return capacity / float(post)
        elif pre in {"D", "-C"}:
            return -1 * capacity / float(post)
    else:
        if val == "C":
            pre = 1.0
        elif val == "D":
            pre = -1.0
        elif "D" in val:
            pre = float(val.replace("D", "")) * -1
        elif "C" in val:
            pre = float(val.replace("C", ""))
        else:
            pre = float(val)
        return pre * capacity


def current(val: Union[list, str, float], capacity: float) -> Union[list[float], float]:
    if isinstance(val, list):
        ret = []
        for v in val:
            ret.append(_current(v, capacity))
    else:
        ret = _current(val, capacity)
    return ret


def vlimit(cond: str, vals: list[float], Is: list[float] = None) -> list[float]:
    if Is is not None:
        signs = [i > 0 for i in Is]
        for i, v in enumerate(vals):
            if cond == "max" and not signs[i]:
                vals[i] = v + 0.0005
            elif cond == "min" and signs[i]:
                vals[i] = v - 0.0005
    return vals


def translate(technique: dict, capacity: float) -> dict:
    if technique["name"].startswith("constant_"):
        ns = get_num_steps(technique)
        tech = {
            "name": technique["name"],
            "Step_number": ns - 1,
            "N_Cycles": technique.get("n_cycles", 0),
            "Record_every_dT": technique.get("record_every_dt", 30.0),
            "I_Range": I_ranges[technique.get("I_range", "keep")],
            "E_Range": E_ranges[technique.get("E_range", "auto")],
            "Duration_step": pad_steps(technique["time"], ns),
            "vs_initial": pad_steps(technique.get("is_delta", False), ns),
            "Test1_Config": pad_steps(0, ns),
            "Test1_Value": pad_steps(0.0, ns),
            "Test2_Config": pad_steps(0, ns),
            "Test2_Value": pad_steps(0.0, ns),
            "Test3_Config": pad_steps(0, ns),
            "Test3_Value": pad_steps(0.0, ns),
            "Exit_Cond": pad_steps(2 * int(technique.get("exit_on_limit", False)), ns),
        }
        ci = 1
        if technique["name"].endswith("current"):
            I = current(technique["current"], capacity)
            tech["Current_step"] = pad_steps(I, ns)
            tech["Record_every_dE"] = technique.get("record_every_dE", 0.005)
        elif technique["name"].endswith("voltage"):
            tech["Voltage_step"] = pad_steps(technique["voltage"], ns)
            tech["Record_every_dI"] = technique.get("record_every_dI", 0.001)
        for prop in {"voltage", "current"}:
            for cond in {"max", "min"}:
                if f"limit_{prop}_{cond}" in technique and ci < 4:
                    conf = get_test_magic(prop, cond)
                    tech[f"Test{ci}_Config"] = pad_steps(conf, ns)
                    if prop == "current":
                        Is = current(technique[f"limit_{prop}_{cond}"], capacity)
                        vals = pad_steps(Is, ns)
                    else:
                        padded = pad_steps(technique[f"limit_{prop}_{cond}"], ns)
                        vals = vlimit(cond, padded, tech.get("Current_step"))
                    tech[f"Test{ci}_Value"] = vals
                    ci += 1
    elif technique["name"] == "loop":
        tech = {
            "name": "loop",
            "loop_N_times": technique.get("n_gotos", -1),
            "protocol_number": technique.get("goto", 0),
        }
    elif technique["name"].startswith("sweep_"):
        ns = get_num_steps(technique)
        tech = {
            "name": technique["name"],
            "Scan_number": ns - 1,
            "N_Cycles": technique.get("n_cycles", 0),
            "I_Range": I_ranges[technique.get("I_range", "keep")],
            "E_Range": E_ranges[technique.get("E_range", "auto")],
            "vs_initial": pad_steps(technique.get("is_delta", False), ns),
            "Scan_Rate": pad_steps(technique.get("scan_rate", 0.001), ns),
            "Test1_Config": pad_steps(0, ns),
            "Test1_Value": pad_steps(0.0, ns),
            "Test2_Config": pad_steps(0, ns),
            "Test2_Value": pad_steps(0.0, ns),
            "Test3_Config": pad_steps(0, ns),
            "Test3_Value": pad_steps(0.0, ns),
            "Exit_Cond": pad_steps(2 * int(technique.get("exit_on_limit", False)), ns),
        }
        ci = 1
        for prop in {"voltage", "current"}:
            for cond in {"max", "min"}:
                if f"limit_{prop}_{cond}" in technique and ci < 4:
                    conf = get_test_magic(prop, cond)
                    val = current(technique[f"limit_{prop}_{cond}"], capacity)
                    tech[f"Test{ci}_Config"] = pad_steps(conf, ns)
                    tech[f"Test{ci}_Value"] = pad_steps(val, ns)
                    ci += 1
        if technique["name"].endswith("current"):
            I = current(technique["current"], capacity)
            tech["Current_step"] = pad_steps(I, ns)
            tech["Begin_measuring_E"] = technique.get("scan_start", 0.0)
            tech["End_measuring_E"] = technique.get("scan_end", 1.0)
            tech["Record_every_dI"] = technique.get("record_every_dI", 0.001)
        elif technique["name"].endswith("voltage"):
            tech["Voltage_step"] = pad_steps(technique["voltage"], ns)
            tech["Begin_measuring_I"] = technique.get("scan_start", 0.0)
            tech["End_measuring_I"] = technique.get("scan_end", 1.0)
            tech["Record_every_dE"] = technique.get("record_every_dE", 0.005)
    else:
        if technique["name"] != "open_circuit_voltage":
            log.error(f"technique name '{technique['name']}' not understood.")
        tech = {
            "name": "open_circuit_voltage",
            "Rest_time_T": technique.get("time", 0.0),
            "Record_every_dT": technique.get("record_every_dt", 30.0),
            "Record_every_dE": technique.get("record_every_dE", 0.005),
            "I_Range": I_ranges[technique.get("I_range", "keep")],
            "E_Range": E_ranges[technique.get("E_range", "auto")],
        }
    return tech


def dsl_to_ecc(api, dsl: dict) -> EccParams:
    eccs = []
    for k, val in dsl.items():
        if k == "name":
            continue
        elif isinstance(val, list):
            for i, v in zip(range(len(val)), val):
                ecc = make_ecc_parm(api, ECC_parm(k, named_params[k]), v, i)
                eccs.append(ecc)
        else:
            ecc = make_ecc_parm(api, ECC_parm(k, named_params[k]), val)
            eccs.append(ecc)
    eccpars = make_ecc_parms(api, *eccs)
    return eccpars


def payload_to_ecc(api, payload: list[dict], capacity: float) -> list[dict]:
    eccs = []
    for technique in payload:
        dsl = translate(technique, capacity)
        eccpars = dsl_to_ecc(api, dsl)
        eccs.append((dsl["name"], eccpars))
    return eccs


def parse_raw_data(api, data, devname):
    current_values, data_info, data_record = data

    status = PROG_STATE(current_values.State).name
    techname = TECH_ID(data_info.TechniqueID).name

    parsed = {
        "technique": {
            "index": data_info.TechniqueIndex,
            "name": techname,
            "data_rows": data_info.NbRows,
            "data_cols": data_info.NbCols,
            "start_time": data_info.StartTime,
            "loop_number": data_info.loop,
        },
        "current": {
            "status": status,
            "elapsed_time": current_values.ElapsedTime,
            "I_range": {v: k for k, v in I_ranges.items()}[current_values.IRange],
            "E_range": {
                "min": current_values.EweRangeMin,
                "max": current_values.EweRangeMax,
            },
            "timebase": current_values.TimeBase,
        },
        "data": [],
    }

    vmp3 = devname in VMP3_FAMILY
    parsed_data = []
    if data_info.NbRows > 0:
        dtypes = datatypes["VMP3" if vmp3 else "SP-300"][techname]

    ix = 0
    for _ in range(data_info.NbRows):
        inx = ix + data_info.NbCols
        point = {}
        t_rel = 0
        for k, v in zip(dtypes, data_record[ix:inx]):
            if k == "t_low":
                t_rel += v
            elif k == "t_high":
                t_rel += v << 32
            elif k == "cycle":
                point[k] = v
            else:
                point[k] = api.ConvertNumericIntoSingle(v)
        point["time"] = t_rel * current_values.TimeBase
        parsed_data.append(point)
        ix = inx
    parsed["data"] = parsed_data

    return parsed


def get_kbio_api(dllpath):
    if c_is_64b:
        dllfile = "EClib64.dll"
    else:
        dllfile = "EClib.dll"
    apipath = os.path.join(dllpath, dllfile)
    log.debug(f"biologic library path is '{apipath}'")
    api = KBIO_api(apipath)
    return api


def get_kbio_techpath(
    dllpath,
    techname,
    devname,
) -> str:
    vmp3 = devname in VMP3_FAMILY
    techfile = techfiles["VMP3" if vmp3 else "SP-300"][techname]
    return os.path.join(dllpath, techfile)
